use crate::indexer::*;
use crate::Api;

use anyhow::{anyhow, Result};
use std::collections::{hash_map, HashMap};

pub struct IdentityCache<'a> {
    api: &'a Api,
    identities: HashMap<u32, String>,
    indexer: Option<(&'a reqwest::Client, &'a str)>,
}

impl<'a> IdentityCache<'a> {
    pub fn new(api: &'a Api, indexer: Option<(&'a reqwest::Client, &'a str)>) -> Self {
        Self {
            api,
            identities: HashMap::new(),
            indexer,
        }
    }

    pub async fn fetch_identity(
        &mut self,
        identity_id: u32,
        parent_hash: sp_core::H256,
    ) -> Result<String> {
        Ok(match self.identities.entry(identity_id) {
            hash_map::Entry::Occupied(entry) => entry.get().clone(),
            hash_map::Entry::Vacant(entry) => entry
                .insert({
                    let pubkey = self
                        .api
                        .storage()
                        .identity()
                        .identities(&identity_id, Some(parent_hash))
                        .await?
                        .ok_or_else(|| anyhow!("Identity {} not found", identity_id))?
                        .owner_key
                        .to_string();
                    if let Some((gql_client, gql_url)) = self.indexer {
                        if let Ok(resp) = post_graphql::<IdentityNameByPubkey, _>(
                            gql_client,
                            gql_url,
                            identity_name_by_pubkey::Variables {
                                pubkey: pubkey.clone(),
                            },
                        )
                        .await
                        {
                            if let Some(data) = resp.data {
                                if let Some(identity) = data.identity_by_pk {
                                    if let Some(name) = identity.name {
                                        format!("“ {} ”", name)
                                    } else {
                                        pubkey
                                    }
                                } else {
                                    pubkey
                                }
                            } else {
                                pubkey
                            }
                        } else {
                            pubkey
                        }
                    } else {
                        pubkey
                    }
                })
                .clone(),
        })
    }
}
